//
//  AmityChatImageTableViewCell.swift
//  SampleApp
//
//  Created by Federico Zanetello on 5/16/18.
//  Copyright © 2018 David Zhang. All rights reserved.
//

import UIKit
import AmitySDK

final class AmityChatImageTableViewCell: UITableViewCell, AmityChatCellProtocol {
    @IBOutlet weak var myImageView: UIImageView!
    @IBOutlet weak var displayNameLabel: UILabel!
    @IBOutlet weak var timeLabel: UILabel!
    @IBOutlet weak var messageStatusImageView: UIImageView!

    private var messageId: String?

    var aspectRatioConstraint: NSLayoutConstraint? {
        didSet {
            if let oldValue = oldValue {
                oldValue.isActive = false
            }
            if let aspectRatioConstraint = aspectRatioConstraint {
                aspectRatioConstraint.isActive = true
            }
        }
    }
    
    var messageRepo: AmityMessageRepository!
    
    var fileRepo = AmityFileRepository(client: AmityManager.shared.client!)
    
    private func set(image: UIImage) {
        myImageView.image = image
    }

    func display(_ message: AmityMessage, client: AmityClient) {
        messageId = message.messageId
        timeLabel.accessibilityIdentifier = "chat-image-cell-attribute-label"

        if message.isDeleted {
            timeLabel.text = "Message Deleted"
            displayNameLabel.text = "Message Deleted"
        } else {
            let flagCount = message.flagCount
            
            displayNameLabel.text = message.user?.displayName ?? "" + " F: (\(flagCount)"
            messageStatusImageView.image = message.syncState.symbolImage()
            messageStatusImageView.accessibilityIdentifier = "chat.message-cell.statusImageView-\(message.syncState)"
            
            setMetadata(for: message)
            
            messageRepo = AmityMessageRepository(client: client)

            guard !message.isDeleted else {
                self.aspectRatioConstraint = self.myImageView.widthAnchor.constraint(equalTo: self.myImageView.heightAnchor,
                multiplier: 1)
                self.myImageView.image = UIImage(named:"picture")
                self.myImageView.tintColor = UIColor.gray
                return
            }
                        
            if message.messageType == .image {
                
                if message.syncState == .synced {
                    // For synced message, the local file path previously returned will be replaced by the actual web url of the image.
                    // You can fetch the image from server and cache it with messageId or fileId as the key.
                    // Or you can still use local image here if its still available.
                    guard let imageInfo = message.getImageInfo() else { return }
                    ImageMessageHandler.shared.fetchImage(fileURL: imageInfo.fileURL) { [weak self] (image) in
                        self?.updateImage(image: image)
                    }
                } else {
                    guard let imageInfo = message.getImageInfo(), let imageURL = URL(string: imageInfo.fileURL) else { return }
                    if let image = UIImage(contentsOfFile: imageURL.path) {
                        Log.add(info: "[IMAGE] Success loading image in syncing state")
                        self.updateImage(image: image)
                    } else {
                        Log.add(info: "[IMAGE] Error loading image in syncing state")
                    }
                    
                    
                }
            }
        }
    }
    
    func updateImage(image: UIImage?) {
        guard let image = image else { return }
        
        let aspect = image.size.width / image.size.height
        self.aspectRatioConstraint = self.myImageView.widthAnchor.constraint(equalTo: self.myImageView.heightAnchor, multiplier: aspect)
        self.myImageView.image = image
    }
    
    private func setMetadata(for message: AmityMessage) {
        let messageCreateDate: NSAttributedString? = createDate(for: message)
        let messageTags: NSAttributedString? = tags(for: message)
        let childMessages: NSAttributedString? = self.childMessages(for: message)
        let hasParent: NSAttributedString? = parent(for: message)
        let readCount: NSAttributedString? = readCount(for: message)
        let deliveredCount: NSAttributedString? = deliveredCount(for: message)

        let mutableAttributedString = NSMutableAttributedString()

        let optionalAttributedStrings: [NSAttributedString?] = [
            hasParent,
            childMessages,
            messageTags,
            messageCreateDate,
            readCount,
            deliveredCount
        ]

        let attributedStrings: [NSAttributedString] = optionalAttributedStrings.compactMap { $0 }
        for attributedString in attributedStrings {
            mutableAttributedString.append(NSAttributedString(string: " "))
            mutableAttributedString.append(attributedString)
        }
        timeLabel.attributedText = mutableAttributedString
    }

    private func createDate(for message: AmityMessage) -> NSAttributedString {
        let createDateString: String = DateFormatter.localizedString(from: message.createdAt,
                                                                     dateStyle: .none,
                                                                     timeStyle: .short)
        return NSAttributedString(string: createDateString)
    }

    private func tags(for message: AmityMessage) -> NSAttributedString? {
        if let messageTags = message.tags as? [String],
            !messageTags.isEmpty {
            return attributedString(for: messageTags)
        }
        return nil
    }

    private func attributedString(for tags: [String]) -> NSAttributedString {
        assert(!tags.isEmpty)
        let colors: [UIColor] = ["PastelRed", "PastelOrange", "PastelYellow",
                                 "PastelGreen", "PastelBlue", "PastelPurple",
                                 "NeonPink", "NeonRed", "NeonGreen", "NeonAzure"].compactMap(UIColor.init)

        let mutableAttributedString = NSMutableAttributedString()
        for index in 0..<tags.count {
            let index = index % colors.count
            let attributedTagString = NSAttributedString(string: " \(tags[index]) ",
                attributes: [.backgroundColor: colors[index],
                             .foregroundColor: .white])
            mutableAttributedString.append(attributedTagString)
            mutableAttributedString.append(NSAttributedString(string: " "))
        }
        return mutableAttributedString
    }

    private func childMessages(for message: AmityMessage) -> NSAttributedString? {
        if message.childrenNumber > 0 {
            let attributedTagString = NSAttributedString(string: " childs: \(message.childrenNumber) ",
                attributes: [.backgroundColor: UIColor.black,
                             .foregroundColor: .white])
            return attributedTagString
        }
        return nil
    }

    private func parent(for message: AmityMessage) -> NSAttributedString? {
        if message.parentId != nil {
            return NSAttributedString(string: "🧒🏻")
        }
        return nil
    }
    
    private func readCount(for message: AmityMessage) -> NSAttributedString? {
        if message.readCount != 0 {
            return NSAttributedString(string: "R: \(message.readCount)")
        }
        return nil
    }
    
    private func deliveredCount(for message: AmityMessage) -> NSAttributedString? {
        if message.deliveredCount != 0 {
            return NSAttributedString(string: "D: \(message.deliveredCount)")
        }
        return nil
    }

    override func prepareForReuse() {
        super.prepareForReuse()
        aspectRatioConstraint = nil
        myImageView.image = nil
        timeLabel.text = ""
        messageStatusImageView.image = nil
    }
}
