//
//  ReadDeliverUserListView.swift
//  SampleApp
//
//  Created by Manuchet Rungraksa on 10/4/2566 BE.
//  Copyright © 2566 BE David Zhang. All rights reserved.
//

import SwiftUI
import AmitySDK

struct ReadDeliverUserListView: View {
    
    @ObservedObject var viewModel: ReadDeliverUserListViewModel
    
    @State var readMemberFilter: [ReadMemberFilterOptions] = [ReadMemberFilterOptions(name: .member),
                                                              ReadMemberFilterOptions(name: .banned),
                                                              ReadMemberFilterOptions(name: .deleted),
                                                              ReadMemberFilterOptions(name: .muted),
                                                              ReadMemberFilterOptions(name: .nonMember)]
    
    @State var deliveredMemberFilter: [DeliveredMemberFilterOptions] = [DeliveredMemberFilterOptions(name: .member),
                                                                      DeliveredMemberFilterOptions(name: .banned),
                                                                      DeliveredMemberFilterOptions(name: .deleted),
                                                                      DeliveredMemberFilterOptions(name: .muted),
                                                                      DeliveredMemberFilterOptions(name: .nonMember)]
    
    @State private var isReadUser = true
    
    var body: some View{
        VStack {
            HStack {
                Text("Type").font(.caption)
                Spacer()
                Picker("isDeleted", selection: $isReadUser) {
                    Text("Read users").tag(true)
                    Text("Delivered users").tag(false)
                    
                }.pickerStyle(.segmented)
            }
            Divider()
            
            ScrollView(.horizontal) {
                HStack {
                    if isReadUser {
                        ForEach(0..<readMemberFilter.count) { index in
                            
                            Button(action: {
                                readMemberFilter[index].isSelected = readMemberFilter[index].isSelected ? false : true
                            }) {
                                VStack {
                                    if readMemberFilter[index].isSelected {
                                        Image(systemName: "checkmark.circle.fill")
                                            .foregroundColor(.green)
                                            .animation(.easeIn, value: 0)
                                    } else {
                                        Image(systemName: "circle")
                                            .foregroundColor(.primary)
                                            .animation(.easeOut, value: 0)
                                    }
                                    Text(readMemberFilter[index].name.rawValue)
                                }
                            }.buttonStyle(BorderlessButtonStyle())
                        }
                    } else {
                        ForEach(0..<deliveredMemberFilter.count) { index in
                            
                            Button(action: {
                                deliveredMemberFilter[index].isSelected = deliveredMemberFilter[index].isSelected ? false : true
                            }) {
                                VStack {
                                    if deliveredMemberFilter[index].isSelected {
                                        Image(systemName: "checkmark.circle.fill")
                                            .foregroundColor(.green)
                                            .animation(.easeIn, value: 0)
                                    } else {
                                        Image(systemName: "circle")
                                            .foregroundColor(.primary)
                                            .animation(.easeOut, value: 0)
                                    }
                                    Text(deliveredMemberFilter[index].name.rawValue)
                                }
                            }.buttonStyle(BorderlessButtonStyle())
                        }
                    }
                }
            }
            
            Divider()
            
            Button(action: {
                
                viewModel.readMemberFilter = readMemberFilter
                viewModel.deliveredMemberFilter = deliveredMemberFilter
                viewModel.queryUsers(isReadUser: isReadUser)
            }){
                Text("Query user list")
                    .frame(alignment: .center)
            }.buttonStyle(BorderlessButtonStyle())
            
            Divider()
            
        }.padding([.leading, .trailing], 30)
            .padding(.top, 15)
        
        List(viewModel.users) { user in
            VStack(alignment: .leading) {
                Text("User Id: \(user.userId)").frame(alignment: .leadingFirstTextBaseline)
                Text("DisplayName: \(user.displayName)")
                Text("isDeleted \(user.isDeleted)" as String)
            }.onAppear {
                if self.viewModel.users.last?.userId == user.userId {
                    self.viewModel.fetchNextPage()
                }
            }
        }
    }
    
}

struct ReadMemberFilterOptions {
    var name: MessageReadMembershipFilter
    var isSelected: Bool = false
}

struct DeliveredMemberFilterOptions {
    var name: MessageDeliveredMembershipFilter
    var isSelected: Bool = false
}

struct ReadDeliverUserListView_Previews: PreviewProvider {
    
    static var previews: some View {
        Text("Read/Deliver user view")
    }
    
}

struct ReadDeliverUserListModel: Identifiable {
    var id: String
    
    let userObject: AmityUser
    
    let userId: String
    let displayName: String
    let isDeleted: Bool

    init(user: AmityUser) {
        userObject = user
        id = user.userId
        userId = user.userId
        displayName = user.displayName ?? ""
        isDeleted = user.isDeleted
    }
    
}


// Concrete implementor
class ReadDeliverUserListViewModel: ObservableObject {
    
    var pageTitle: String
    var message: AmityMessage
        
    private var token: AmityNotificationToken?
    private var usersCollection: AmityCollection<AmityUser>?
        
    @Published var users: [ReadDeliverUserListModel] = []
    
    var readMemberFilter: [ReadMemberFilterOptions]?
    var deliveredMemberFilter: [DeliveredMemberFilterOptions]?
    
    init(message: AmityMessage) {
        self.pageTitle = "User list"
        self.message = message
    }
    
    func queryUsers(isReadUser: Bool) {
        users.removeAll()
        token?.invalidate()

        if isReadUser {
            let memberFilter = readMemberFilter?.filter({$0.isSelected}).map({$0.name})
            var memberships: Set<MessageReadMembershipFilter>?
            if let memberFilter, memberFilter.count != 0 {
                memberships = Set(memberFilter)
            }
            usersCollection = message.getReadUsers(memberships: memberships)
            token = usersCollection?.observe({ [weak self] (collection, _, _) in
                guard let self = self else { return }
                var list = [ReadDeliverUserListModel]()
                for user in collection.allObjects() {
                    let model = ReadDeliverUserListModel(user: user)
                    list.append(model)
                }
                self.users = list
            })
        } else {
            let memberFilter = deliveredMemberFilter?.filter({$0.isSelected}).map({$0.name})
            var memberships: Set<MessageDeliveredMembershipFilter>?
            if let memberFilter, memberFilter.count != 0 {
                memberships = Set(memberFilter)
            }
            usersCollection = message.getDeliveredUsers(memberships: memberships)
            token = usersCollection?.observe({ [weak self] (collection, _, _) in

                guard let self = self else { return }
                var list = [ReadDeliverUserListModel]()
                for user in collection.allObjects() {
                    let model = ReadDeliverUserListModel(user: user)
                    list.append(model)
                }
                self.users = list
            })
        }
    }
    
    func fetchNextPage() {
        guard usersCollection?.hasNext ?? false else { return }
        usersCollection?.nextPage()
    }
    
}
