//
//  manager.swift
//  SampleApp
//
//  Created by Nishan Niraula on 2/2/22.
//  Copyright © 2022 David Zhang. All rights reserved.
//

import Foundation
import AmitySDK

enum FollowListType {
    case myFollowing(status: AmityFollowQueryOption)
    case myFollower(status: AmityFollowQueryOption)
    case userFollowing(userId: String)
    case userFollower(userId: String)
    
    var isForCurrentUser: Bool {
        switch self {
        case .myFollowing, .myFollower:
            return true
        default:
            return false
        }
    }
}

protocol FollowManagerDelegate: AnyObject {
    func dataDidChange()
}

class FollowManager {
    
    private let userRepo: AmityUserRepository
    private var userRelationship: AmityUserRelationship {
        return userRepo.userRelationship
    }
    private var followCollection: AmityCollection<AmityFollowRelationship>?
    private var token: AmityNotificationToken?
    
    var topicSubscriber: AmityTopicSubscription
    
    weak var delegate: FollowManagerDelegate?
    
    private(set) var follows: [AmityFollowRelationship] = []
    
    init() {
        userRepo = AmityUserRepository(client: AmityManager.shared.client!)
        topicSubscriber = AmityTopicSubscription(client: AmityManager.shared.client!)
    }
    
    var type: FollowListType = .myFollower(status: .pending)
    
    func reloadData() {
        setup()
    }
    
    func nextPage() {
        if let followCollection = followCollection, followCollection.loadingStatus == .loaded {
            followCollection.nextPage()
        }
    }
    
    private func setup() {
        token?.invalidate()
        
        switch type {
        case .myFollowing(let status):
            followCollection = userRelationship.getMyFollowings(with: status)
        case .myFollower(let status):
            followCollection = userRelationship.getMyFollowers(with: status)
        case .userFollowing(let userId):
            followCollection = userRelationship.getFollowings(withUserId: userId)
        case .userFollower(let userId):
            followCollection = userRelationship.getFollowers(withUserId: userId)
        }
        
        token = followCollection?.observe { [weak self] (collection, _, error) in
            var follows: [AmityFollowRelationship] = []
            for follow in collection.allObjects() {
                follows.append(follow)
            }
            self?.follows = follows
            self?.delegate?.dataDidChange()
        }
    }
    
    var followInfoToken: AmityNotificationToken?
    
    func getMyFollowInfo(completion: ((Result<AmityMyFollowInfo, Error>) -> Void)?) {
        followInfoToken?.invalidate()
        followInfoToken = userRelationship.getMyFollowInfo().observe { [weak self] liveObject, error in
            guard let object = liveObject.object, liveObject.dataStatus == .fresh else {
                return
            }
            self?.followInfoToken?.invalidate()
            completion?(.success(object))
        }
    }
    
    func getUserFollowInfo(userId: String, completion: ((Result<AmityUserFollowInfo, Error>) -> Void)?) {
        followInfoToken?.invalidate()
        followInfoToken = userRelationship.getFollowInfo(withUserId: userId).observe({ [weak self] liveObject, error in
            guard let object = liveObject.object, liveObject.dataStatus == .fresh else {
                return
            }
            self?.followInfoToken?.invalidate()
            completion?(.success(object))
        })
    }
    
    func acceptUserRequest(userId: String) {
        Task { @MainActor in
            do {
                let result = try await userRelationship.acceptMyFollower(withUserId: userId)
                print("-> accept \(result.0 ? "success" : "fail")")
            } catch {
                print("-> accept fail")
            }
        }
    }
    
    func declineUserRequest(userId: String) {
        Task { @MainActor in
            do {
                let result = try await userRelationship.declineMyFollower(withUserId: userId)
                print("-> decline \(result.0 ? "success" : "fail")")
            } catch {
                print("-> decline fail")
            }
        }
    }
    
    func unfollowUser(userId: String) {
        Task { @MainActor in
            do {
                let success = try await userRelationship.unfollow(withUserId: userId)
                print("-> unfollow request \(success.0 ? "success" : "fail")")
            } catch {
                print("-> cancel request fail")
            }
        }
    }
    
    func followUser(userId: String) {
        Task { @MainActor in
            do {
                let success = try await userRelationship.follow(withUserId: userId)
                print("-> follow request \(success.0 ? "success" : "fail")")
            } catch {
                print("-> cancel request fail")
            }
        }
    }
    
    func subscribeEvent(event: AmityFollowEvent, completion: ((Bool) -> Void)?) {
        guard type.isForCurrentUser else { return }

        let topic = AmityFollowTopic(event: event)
        topicSubscriber.subscribeTopic(topic) { isSuccess, error in
            completion?(isSuccess)
            Log.add(info: "Follow Topic subscribe: \(isSuccess) Error: \(String(describing: error))")
        }
    }

    func unsubscribeEvent(event: AmityFollowEvent, completion: ((Bool) -> Void)?) {
        guard type.isForCurrentUser else { return }

        let topic = AmityFollowTopic(event: event)
        topicSubscriber.unsubscribeTopic(topic) { isSuccess, error in
            completion?(isSuccess)
            Log.add(info: "Follow Topic unsubscribe: \(isSuccess) Error: \(String(describing: error))")
        }
    }
}
