//
//  UserListViewController.swift
//  SampleApp
//
//  Created by Nishan Niraula on 5/11/20.
//  Copyright © 2020 David Zhang. All rights reserved.
//

import UIKit
import AmitySDK

class UserListViewController: UIViewController, UISearchBarDelegate {

    @IBOutlet weak var tableView: UITableView!
    @IBOutlet weak var searchBar: UISearchBar!
    @IBOutlet weak var sortBySegment: UISegmentedControl!
    
    let feedbackImpactGenerator = UIImpactFeedbackGenerator()
    
    var listManager: UserListManager!
    var userRepository: AmityUserRepository!
    var userRelationship: AmityUserRelationship!
    
    var userListToken: AmityNotificationToken?
    var searchText: String?
    var sortBy: AmityUserSortOption = .displayName
    
    override func viewDidLoad() {
        super.viewDidLoad()
        
        listManager = UserListManager(client: AmityManager.shared.client!)
        userRepository = AmityUserRepository(client: AmityManager.shared.client!)
        setupViews()
    }
    
    override func viewWillAppear(_ animated: Bool) {
        super.viewWillAppear(animated)
        
        fetchAllUsetList()
    }
    
    override func viewWillDisappear(_ animated: Bool) {
        super.viewWillDisappear(animated)
        
        listManager.presenceRepository.unsyncAllUserPresence()
    }
    
    func updateUserList() {
        self.tableView.reloadData()
    }
    
    func setupViews() {
        self.title = "Users"
        
        tableView.dataSource = self
        tableView.delegate = self
        tableView.reloadData()
        tableView.keyboardDismissMode = .onDrag
        
        searchBar.searchBarStyle = .minimal
        searchBar.placeholder = "Search user name"
        searchBar.delegate = self
        
        let addDescriptionButton = UIBarButtonItem(barButtonSystemItem: .add, target: self, action: #selector(addDescription))
        self.navigationItem.rightBarButtonItem = addDescriptionButton
    }
    
    func searchBar(_ searchBar: UISearchBar, textDidChange searchText: String) {
        self.searchText = searchText
        self.listManager.presenceRepository.unsyncAllUserPresence() // Unsync all
        listManager.searchUserList(name: searchText, sortBy: sortBy) { [weak self] in
            self?.updateUserList()
        }
    }
    
    @IBAction func sortBySegmentAction(_ sender: Any) {
        if let segment = sender as? UISegmentedControl {
            switch segment.selectedSegmentIndex {
            case 0:
                sortBy = .displayName
            case 1:
                sortBy = .firstCreated
            case 2:
                sortBy = .lastCreated
            default:
                break
            }
        }
        
        guard let text = searchText else {
            fetchAllUsetList()
            return
        }
        
        listManager.searchUserList(name: text, sortBy: sortBy) { [weak self] in
            self?.updateUserList()
        }
    }
    
    func fetchAllUsetList() {
        listManager.fetchUserList(sortedBy: sortBy) { [weak self] in
            self?.updateUserList()
        }
    }
    
    func scrollViewWillEndDragging(_ scrollView: UIScrollView, withVelocity velocity: CGPoint, targetContentOffset: UnsafeMutablePointer<CGPoint>) {
        
        let frameHeight = scrollView.frame.size.height
        let contentOffset = scrollView.contentOffset.y
        let contentHeight = scrollView.contentSize.height
        
        let distanceFromBottom = contentHeight - contentOffset
        if distanceFromBottom < frameHeight {
            listManager.loadMoreUsers()
        }
    }
    
    func displayUserFeed(userId: String, userName: String?) {
        let feedManager = UserPostsFeedManager(client: AmityManager.shared.client!, userId: userId, userName: userName)
        feedManager.feedType = .userFeed
        
        let postsFeedStoryboard = UIStoryboard(name: "Feed", bundle: nil)
        let postsFeedController = postsFeedStoryboard.instantiateViewController(withIdentifier: UserPostsFeedViewController.identifier) as! UserPostsFeedViewController
        postsFeedController.feedManager = feedManager
        
        self.navigationController?.pushViewController(postsFeedController, animated: true)
    }
    
    @objc func addDescription() {
        UIAlertController.showAlertForUserInput(in: self, title: "User", message: "Set user description:") { description in
            UserUpdateManager.shared.updateDescription(description: description, completion: nil)
        }
    }
}

extension UserListViewController: UITableViewDataSource, UITableViewDelegate {
    
    func tableView(_ tableView: UITableView, numberOfRowsInSection section: Int) -> Int {
        return listManager.numberOfUsers()
    }
    
    func tableView(_ tableView: UITableView, cellForRowAt indexPath: IndexPath) -> UITableViewCell {
        
        let cell = tableView.dequeueReusableCell(withIdentifier: UserListTableViewCell.identifier) as! UserListTableViewCell
        
        let user = listManager.getUserItem(at: indexPath.row)
        cell.delegate = self
        cell.userNameLabel.text = user?.displayName ?? "-"
        cell.userIdLabel.text = "Id: \(user?.userId ?? "")"
        cell.avatarLabel.text = "Avatar: Id: \(user?.avatarFileId ?? ""), URL: \(user?.avatarCustomUrl ?? "")"
        cell.infoLabel.text = "IsDeleted: \(user?.isDeleted ?? false)"
        cell.heartbeatLabel.text = "Heartbeat: \(user?.lastHeartbeat?.description ?? "")"
        
        if let userHeartbeat = user?.lastHeartbeat, Date().timeIntervalSince(userHeartbeat) < 60 {
            // User is online
            cell.onlineIndicatorView.isHidden = false
        } else {
            // User is offline.
            cell.onlineIndicatorView.isHidden = true
        }
        
        return cell
    }
    
    func tableView(_ tableView: UITableView, didSelectRowAt indexPath: IndexPath) {
        guard let user = listManager.getUserItem(at: indexPath.row) else { return }
        
        var actions = [UIAlertAction]()
        actions.append(UIAlertAction(title: "View User Feed", style: .default, handler: { [weak self] action in
            self?.displayUserFeed(userId: user.userId, userName: user.displayName)
        }))
        
        actions.append(UIAlertAction(title: "Real Time Event", style: .default, handler: { [weak self] action in
            self?.displayRealTimeEventScreen(user: user)
        }))
        
        actions.append(UIAlertAction(title: "Flag User", style: .default, handler: { [weak self] action in
            self?.flagUser(user: user)
        }))
        
        actions.append(UIAlertAction(title: "Unflag User", style: .default, handler: { [weak self] action in
            self?.unflagUser(user: user)
        }))
        
        AppUtility.showActionSheet(in: self, title: "", message: "What do you want to do?", actions: actions)
    }
    
    func tableView(_ tableView: UITableView, willDisplay cell: UITableViewCell, forRowAt indexPath: IndexPath) {
        guard let user = listManager.getUserItem(at: indexPath.row) else { return }
        
        listManager.presenceRepository.syncUserPresence(id: user.userId)
    }
    
    func tableView(_ tableView: UITableView, didEndDisplaying cell: UITableViewCell, forRowAt indexPath: IndexPath) {
        guard let user = listManager.getUserItem(at: indexPath.row), !listManager.isInSearchMode else { return }
        
        listManager.presenceRepository.unsyncUserPresence(id: user.userId)
    }
    
    func displayRealTimeEventScreen(user: AmityUser) {
        let controller = UserRealTimeEventController()
        controller.manager = UserRealTimeEventManager(user: user)
        self.navigationController?.pushViewController(controller, animated: true)
    }
    
    func flagUser(user: AmityUser) {
        isUserFlagged(user: user) { [weak self] isFlagged in
            guard let strongSelf = self else { return }
            
            guard !isFlagged else {
                AppUtility.showAlert(in: strongSelf, title: "", message: "User is previously flagged", action: nil)
                return
            }
            
            Task { @MainActor in
                let message: String
                do {
                    let result = try await strongSelf.userRepository?.flagUser(withId: user.userId)
                    message = "User is flagged successfully"
                } catch {
                    message = "Error: \(String(describing: error.localizedDescription))"
                }
                AppUtility.showAlert(in: strongSelf, title: "", message: message, action: nil)
            }
        }
    }
    
    func unflagUser(user: AmityUser) {
        isUserFlagged(user: user) { [weak self] isFlagged in
            guard let strongSelf = self else { return }
            
            guard isFlagged else {
                AppUtility.showAlert(in: strongSelf, title: "", message: "User is not flagged yet", action: nil)
                return
            }
            
            Task { @MainActor in
                let message: String
                do {
                    let result = try await strongSelf.userRepository?.unflagUser(withId: user.userId)
                    message = "User is unflagged successfully"
                } catch {
                    message = "Error: \(String(describing: error.localizedDescription))"
                }
                AppUtility.showAlert(in: strongSelf, title: "", message: message, action: nil)
            }
        }
    }
    
    func isUserFlagged(user: AmityUser, completion: @escaping (Bool) -> Void) {
        Task { @MainActor in
            do {
                let isFlagged = await try userRepository.isUserFlaggedByMe(withId: user.userId)
                completion(isFlagged)
            } catch {
                completion(false)
            }
        }
    }
    
    func blockUser(userId: String) {
        Task { @MainActor in
            let message: String
            do {
                try await userRelationship.blockUser(userId: userId)
                message = "User \(userId) is blocked"
            } catch let error {
                message = "Error: \(error.localizedDescription)"
            }
            
            AppUtility.showAlert(in: self, title: "", message: message, action: nil)
        }
    }
    
    func unblockUser(userId: String) {
        Task { @MainActor in
            let message: String
            do {
                try await userRelationship.unblockUser(userId: userId)
                message = "User \(userId) is unblocked"
            } catch let error {
                message = "Error: \(error.localizedDescription)"
            }
            
            AppUtility.showAlert(in: self, title: "", message: message, action: nil)
        }
    }
}

enum FollowAction: String {
    case follow
    case unfollow
    case accept
    case decline
}

extension UserListViewController: UserListTableViewCellDelegate {
    
    func cellOptionDidTap(_ cell: UserListTableViewCell) {
        guard let indexPath = tableView.indexPath(for: cell),
              let userId = listManager.getUserItem(at: indexPath.row)?.userId else { return }
        
        feedbackImpactGenerator.impactOccurred()
        
        userRelationship = userRepository.userRelationship
        userListToken?.invalidate()
        
        userListToken = userRelationship.getFollowInfo(withUserId: userId).observeOnce { [weak self] liveObject, error in
            
            guard let info = liveObject.object, liveObject.dataStatus == .fresh, let weakSelf = self else { return }
            
            // Invalidate this token
            self?.userListToken?.invalidate()
            self?.userListToken = nil
            
            let followInfoDesc = "Follow Status: " + info.status.title
            
            var actionsList = [UIAlertAction]()
            
            let followAction = UIAlertAction(title: "Follow", style: .default) { [weak self] _ in
                Task { @MainActor in
                    do {
                        let result = try await self?.userRelationship.follow(withUserId: userId)
                        self?.displayFollowActionAlert(action: .follow, status: result?.0 ?? false, error: nil)
                    } catch {
                        self?.displayFollowActionAlert(action: .follow, status: false, error: error)
                    }
                }
            }
            
            let unfollowAction = UIAlertAction(title: "Unfollow", style: .default) { [weak self] _ in
                Task { @MainActor in
                    do {
                        let result = try await self?.userRelationship.unfollow(withUserId: userId)
                        self?.displayFollowActionAlert(action: .unfollow, status: result?.0 ?? false, error: nil)
                    } catch {
                        self?.displayFollowActionAlert(action: .unfollow, status: false, error: error)
                    }
                }
            }
            
            let blockAction = UIAlertAction(title: "Block", style: .default) { [weak self] _ in
                self?.blockUser(userId: userId)
            }
            
            let unblockAction = UIAlertAction(title: "Unblock", style: .default) { [weak self] _ in
                self?.unblockUser(userId: userId)
            }
            
            switch info.status {
            case .none:
                actionsList.append(followAction)
                actionsList.append(blockAction)
            case .blocked:
                actionsList.append(unblockAction)
            case .accepted:
                actionsList.append(unfollowAction) // unfollow
                actionsList.append(blockAction)
            case .pending:
                actionsList.append(unfollowAction) // withdraw request
                actionsList.append(blockAction)
            @unknown default:
                break
            }

            AppUtility.showActionSheet(in: weakSelf, title: "Actions\n\(followInfoDesc)", message: "", actions: actionsList)
        }
    }
    
    func displayFollowActionAlert(action: FollowAction, status: Bool, error: Error?) {
        let title: String = status == true ? "\(action.rawValue.capitalized) Successful" : "\(action.rawValue.capitalized) Failed"
        let message: String = error == nil ? "" : error?.localizedDescription ?? ""
        
        // Show this alert after 1 second.
        AppUtility.showAlert(in: self, title: title, message: message, action: nil)
    }
}

extension UIAlertController {
    
    class func showAlertForUserInput(in vc: UIViewController, title: String, message: String, action: ((_ input: String) -> Void)?) {
        
        let alertController = UIAlertController(title: title, message: message, preferredStyle: .alert)
        alertController.addTextField(configurationHandler: { textField in
            textField.placeholder = "Input"
        })

        let inputAction = UIAlertAction(title: "Submit", style: .default, handler: { _ in
            guard let userInput = alertController.textFields?.first?.text else { return }
            action?(userInput)
        })
        alertController.addAction(inputAction)

        let cancelAction = UIAlertAction(title: "Cancel", style: .cancel, handler: nil)
        alertController.addAction(cancelAction)

        vc.present(alertController, animated: true, completion: nil)
    }
}
